#=======================================================================
# Formas eficientes para fazer estudos de simulação/reamostragem.

#-----------------------------------------------------------------------
# Cria a função com lista de todos os argumentos e use mapply().

library(microbenchmark)

# Curva de poder.
power_k_groups <- function(k = 3,
                           r = 4,
                           sigma_group = 0.01,
                           simulations = 50,
                           alpha = 0.05) {
    trt <- gl(n = k, k = r)
    X <- model.matrix(~0 + trt)
    results <- replicate(simulations, {
        b <- rnorm(k, mean = 0, sigma_group)
        y <- rnorm(k * r, mean = X %*% b, sd = 1)
        anova(lm(y ~ trt))[1, 5]
    })
    mean(results <= alpha)
}

# Criar o grid de condições (prototipar com um grid pequeno).
grid <- expand.grid(k = c(3, 5, 8),
                    r = c(3, 8),
                    sigma_group = seq(0.01, 2, length.out = 20),
                    KEEP.OUT.ATTRS = FALSE)
nrow(grid)
grid

# Aplicar.
grid$taxa <- mapply(FUN = power_k_groups,
                    k = grid$k,
                    r = grid$r,
                    sigma_group = grid$sigma_group,
                    MoreArgs = list(simulations = 100,
                                    alpha = 0.05),
                    SIMPLIFY = TRUE)

#-----------------------------------------------------------------------

library(latticeExtra)

xyplot(taxa ~ sigma_group | ordered(r),
       groups = k,
       auto.key = TRUE,
       data = grid,
       type = "o")

apropos("power")
apropos("write")

#-----------------------------------------------------------------------
# Usando purrr.

library(purrr)
ls("package:purrr")

grid$taxa <- NULL
grid$taxa <- pmap_dbl(.l = grid,
                      .f = power_k_groups,
                      simulations = 100,
                      alpha = 0.05)

xyplot(taxa ~ sigma_group | ordered(r),
       groups = k,
       auto.key = TRUE,
       data = grid,
       type = "o")

#-----------------------------------------------------------------------

library(microbenchmark)

microbenchmark(
    mapply = {
        mapply(FUN = power_k_groups,
               k = grid$k,
               r = grid$r,
               sigma_group = grid$sigma_group,
               MoreArgs = list(simulations = 10,
                               alpha = 0.05),
               SIMPLIFY = TRUE)
    },
    pmap = {
        pmap_dbl(.l = grid,
                 .f = power_k_groups,
                 simulations = 10,
                 alpha = 0.05)
    },
    replications = 20)

#-----------------------------------------------------------------------
# Mais um exemplo.

library(latticeExtra)

# Situação: Para fazer um classificador por ensemble são usados
# `n_class` classificadores independentes (base/weak/lazy learners) que
# tem taxa de acerto individual de `tx_acer`. Qual é a taxa de acerto do
# ensemble de classificadores considerando como decisão a classe
# prevista por maioria simples? Fazer a curva de taxa de acerto do
# ensemble em função da taxa de acerto individual e do número de
# classificadores base do ensamble, ou seja, determinar
# `acc(n_class, tx_acer)`.

simul <- function(n_class = 3, tx_acer = 0.55, N = 1000) {
    #' @param n_class (integer[1] > 3) é o número de classificadores do
    #'     ensemble.
    #' @param tx_acer (0 < double[1] < 1) é a taxa de acerto individual
    #'     de cada classificador independente do ensamble.
    #' @param N (integer[1] > 1) é o número de simulações Monte Carlo.
    #' @return (double[1]) a taxa de acerto do classificador ensemble.
    u <- replicate(N, {
        # Classe correta.
        y <- rbinom(n = 1, size = 1, prob = 0.5)
        # Classe predita por cada classificador.
        x <- rbinom(n = n_class,
                    size = 1,
                    prob = 1 - abs(y - tx_acer))
        # Classe predita pela coleção usando maioria simples.
        z <- as.integer(mean(x) > 0.5)
        # Verifica se houve acerto.
        z == y
    })
    mean(u)
}

# Comparando teórico com simulado.
k <- 13
alpha <- 0.7
sum(dbinom(ceiling(k/2):k, size = k, prob = alpha))
simul(n_class = k, tx_acer = alpha, N = 10000)

# Grid de valores nos parâmetros da simulação.
grid <- expand.grid(n_class = c(3, 5, 9, 13, 19, 25),
                    tx_acer = seq(0.35, 0.95, by = 0.025))
nrow(grid)

# Obtenção da acurácia do ensemble em cada ponto de suporte.
grid$acc <- with(grid,
                 mapply(FUN = simul,
                        n_class,
                        tx_acer,
                        MoreArgs = list(N = 2000)))

# Visualização das curvas.
xyplot(acc ~ tx_acer,
       groups = n_class,
       data = grid,
       xlab = "Taxa de acerto do base learner",
       ylab = "Taxa de acerto do ensemble",
       auto.key = list(corner = c(0.95, 0.05),
                       title = "#Classif.",
                       cex.title = 1),
       grid = TRUE,
       type = "p") +
    layer({
        panel.abline(v = 0.5, h = 0.5, lty = 2)
    }) +
    glayer({
        k <- group.value
        p <- sapply(x,
                    FUN = function(alpha) {
                        sum(dbinom(ceiling(k/2):k,
                                   size = k,
                                   prob = alpha))
                    })
        panel.lines(x, p, col = col.line)
    })

#-----------------------------------------------------------------------

# Situação: Para fazer um regressor por ensemble são usados `n_regr`
# regressores independentes (base/weak/lazy learners) que tem
# desvio-padrão do erro individual de predição `pr_err`. Qual é o erro
# quadrático médio de predição do ensemble de regressores considerando
# como valor predito a média aritmética das previsões dos regressores?
# Fazer a curva de erro quadrático médio do ensemble em função do erro
# de predição individual e do número de regressores base do ensamble, ou
# seja, determinar `err(n_regr, pr_err)`.

simul <- function(n_regr = 3, pr_err = 1, N = 1000) {
    #' @param n_regr (integer[1] > 3) é o número de regressores do
    #'     ensemble.
    #' @param pr_err (0 < double[1]) é o desvio-padrão do erro de
    #'     predição individual de cada regressor independente do
    #'     ensamble.
    #' @param N (integer[1] > 1) é o número de simulações Monte Carlo.
    #' @return (double[1]) o erro quadrático médio de predição do
    #'     regressor ensemble.
    u <- replicate(N, {
        # Valor correto.
        y <- runif(n = 1, min = -5, max = 5)
        # Valor predito por cada regressor.
        x <- rnorm(n = n_regr,
                   mean = y,
                   sd = pr_err)
        # Valor predito pela coleção de regressores.
        z <- mean(x)
        (z - y)^2
    })
    # Determina o erro quadrático médio.
    mean(u)
}

# Grid de valores nos parâmetros da simulação.
grid <- expand.grid(n_regr = c(3, 5, 9, 13, 19, 25),
                    pr_err = seq(0.1, 10, by = 0.2))
nrow(grid)

# Obtenção da acurácia do ensemble em cada ponto de suporte.
grid$err <- with(grid,
                 mapply(FUN = simul,
                        n_regr,
                        pr_err,
                        MoreArgs = list(N = 2000)))

# Visualização das curvas.
xyplot(err ~ pr_err,
       groups = n_regr,
       data = grid,
       xlab = "Desvio-padrão do erro de predição do base learner",
       ylab = "Erro quadrático médio de predição do ensemble",
       auto.key = list(corner = c(0.05, 0.95),
                       title = "#Regress.",
                       cex.title = 1),
       grid = TRUE,
       type = "p") +
    glayer({
        panel.curve(x^2/group.value, from = 0,
                    lty = 2, col = col.line)
    })

#-----------------------------------------------------------------------
